#ifndef POINTER_GESTURES_CLIENT_PROTOCOL_H
#define POINTER_GESTURES_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct _wl_pointer_gesture_pinch;
struct _wl_pointer_gesture_swipe;
struct _wl_pointer_gestures;
struct wl_pointer;
struct wl_surface;

extern const struct wl_interface _wl_pointer_gestures_interface;
extern const struct wl_interface _wl_pointer_gesture_swipe_interface;
extern const struct wl_interface _wl_pointer_gesture_pinch_interface;

#define _WL_POINTER_GESTURES_GET_SWIPE_GESTURE	0
#define _WL_POINTER_GESTURES_GET_PINCH_GESTURE	1

static inline void
_wl_pointer_gestures_set_user_data(struct _wl_pointer_gestures *_wl_pointer_gestures, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) _wl_pointer_gestures, user_data);
}

static inline void *
_wl_pointer_gestures_get_user_data(struct _wl_pointer_gestures *_wl_pointer_gestures)
{
	return wl_proxy_get_user_data((struct wl_proxy *) _wl_pointer_gestures);
}

static inline void
_wl_pointer_gestures_destroy(struct _wl_pointer_gestures *_wl_pointer_gestures)
{
	wl_proxy_destroy((struct wl_proxy *) _wl_pointer_gestures);
}

static inline struct _wl_pointer_gesture_swipe *
_wl_pointer_gestures_get_swipe_gesture(struct _wl_pointer_gestures *_wl_pointer_gestures, struct wl_pointer *pointer)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) _wl_pointer_gestures,
			 _WL_POINTER_GESTURES_GET_SWIPE_GESTURE, &_wl_pointer_gesture_swipe_interface, NULL, pointer);

	return (struct _wl_pointer_gesture_swipe *) id;
}

static inline struct _wl_pointer_gesture_pinch *
_wl_pointer_gestures_get_pinch_gesture(struct _wl_pointer_gestures *_wl_pointer_gestures, struct wl_pointer *pointer)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) _wl_pointer_gestures,
			 _WL_POINTER_GESTURES_GET_PINCH_GESTURE, &_wl_pointer_gesture_pinch_interface, NULL, pointer);

	return (struct _wl_pointer_gesture_pinch *) id;
}

/**
 * _wl_pointer_gesture_swipe - a swipe gesture object
 * @begin: multi-finger swipe begin
 * @update: multi-finger swipe motion
 * @end: multi-finger swipe end
 *
 * A swipe gesture object notifies a client about a multi-finger swipe
 * gesture detected on an indirect input device such as a touchpad. The
 * gesture is usually initiated by multiple fingers moving in the same
 * direction but once initiated the direction may change. The precise
 * conditions of when such a gesture is detected are
 * implementation-dependent.
 *
 * A gesture consists of three stages: begin, update (optional) and end.
 * There cannot be multiple simultaneous pinch or swipe gestures on a same
 * pointer/seat, how compositors prevent these situations is
 * implementation-dependent.
 *
 * A gesture may be cancelled by the compositor or the hardware. Clients
 * should not consider performing permanent or irreversible actions until
 * the end of a gesture has been received.
 */
struct _wl_pointer_gesture_swipe_listener {
	/**
	 * begin - multi-finger swipe begin
	 * @serial: (none)
	 * @time: timestamp with millisecond granularity
	 * @surface: (none)
	 * @fingers: number of fingers
	 *
	 * This event is sent when a multi-finger swipe gesture is
	 * detected on the device.
	 */
	void (*begin)(void *data,
		      struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe,
		      uint32_t serial,
		      uint32_t time,
		      struct wl_surface *surface,
		      uint32_t fingers);
	/**
	 * update - multi-finger swipe motion
	 * @time: timestamp with millisecond granularity
	 * @dx: delta x coordinate in surface coordinate space
	 * @dy: delta y coordinate in surface coordinate space
	 *
	 * This event is sent when a multi-finger swipe gesture changes
	 * the position of the logical center.
	 *
	 * The dx and dy coordinates are relative coordinates of the
	 * logical center of the gesture compared to the previous event.
	 */
	void (*update)(void *data,
		       struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe,
		       uint32_t time,
		       wl_fixed_t dx,
		       wl_fixed_t dy);
	/**
	 * end - multi-finger swipe end
	 * @serial: (none)
	 * @time: timestamp with millisecond granularity
	 * @cancelled: 1 if the gesture was cancelled, 0 otherwise
	 *
	 * This event is sent when a multi-finger swipe gesture ceases to
	 * be valid. This may happen when one or more finger is lifted or
	 * the gesture is cancelled.
	 *
	 * When a gesture is cancelled, the client should undo state
	 * changes caused by this gesture. What causes a gesture to be
	 * cancelled is implementation-dependent.
	 */
	void (*end)(void *data,
		    struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe,
		    uint32_t serial,
		    uint32_t time,
		    int32_t cancelled);
};

static inline int
_wl_pointer_gesture_swipe_add_listener(struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe,
				       const struct _wl_pointer_gesture_swipe_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) _wl_pointer_gesture_swipe,
				     (void (**)(void)) listener, data);
}

#define _WL_POINTER_GESTURE_SWIPE_DESTROY	0

static inline void
_wl_pointer_gesture_swipe_set_user_data(struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) _wl_pointer_gesture_swipe, user_data);
}

static inline void *
_wl_pointer_gesture_swipe_get_user_data(struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe)
{
	return wl_proxy_get_user_data((struct wl_proxy *) _wl_pointer_gesture_swipe);
}

static inline void
_wl_pointer_gesture_swipe_destroy(struct _wl_pointer_gesture_swipe *_wl_pointer_gesture_swipe)
{
	wl_proxy_marshal((struct wl_proxy *) _wl_pointer_gesture_swipe,
			 _WL_POINTER_GESTURE_SWIPE_DESTROY);

	wl_proxy_destroy((struct wl_proxy *) _wl_pointer_gesture_swipe);
}

/**
 * _wl_pointer_gesture_pinch - a pinch gesture object
 * @begin: multi-finger pinch begin
 * @update: multi-finger pinch motion
 * @end: multi-finger pinch end
 *
 * A pinch gesture object notifies a client about a multi-finger pinch
 * gesture detected on an indirect input device such as a touchpad. The
 * gesture is usually initiated by multiple fingers moving towards each
 * other or away from each other, or by two or more fingers rotating around
 * a logical center of gravity. The precise conditions of when such a
 * gesture is detected are implementation-dependent.
 *
 * A gesture consists of three stages: begin, update (optional) and end.
 * There cannot be multiple simultaneous pinch or swipe gestures on a same
 * pointer/seat, how compositors prevent these situations is
 * implementation-dependent.
 *
 * A gesture may be cancelled by the compositor or the hardware. Clients
 * should not consider performing permanent or irreversible actions until
 * the end of a gesture has been received.
 */
struct _wl_pointer_gesture_pinch_listener {
	/**
	 * begin - multi-finger pinch begin
	 * @serial: (none)
	 * @time: timestamp with millisecond granularity
	 * @surface: (none)
	 * @fingers: number of fingers
	 *
	 * This event is sent when a multi-finger pinch gesture is
	 * detected on the device.
	 */
	void (*begin)(void *data,
		      struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch,
		      uint32_t serial,
		      uint32_t time,
		      struct wl_surface *surface,
		      uint32_t fingers);
	/**
	 * update - multi-finger pinch motion
	 * @time: timestamp with millisecond granularity
	 * @dx: delta x coordinate in surface coordinate space
	 * @dy: delta y coordinate in surface coordinate space
	 * @scale: scale relative to the initial finger position
	 * @rotation: angle in degrees cw relative to the previous event
	 *
	 * This event is sent when a multi-finger pinch gesture changes
	 * the position of the logical center, the rotation or the relative
	 * scale.
	 *
	 * The dx and dy coordinates are relative coordinates in the
	 * surface coordinate space of the logical center of the gesture.
	 *
	 * The scale factor is an absolute scale compared to the
	 * pointer_gesture_pinch.begin event, e.g. a scale of 2 means the
	 * fingers are now twice as far apart as on
	 * pointer_gesture_pinch.begin.
	 *
	 * The rotation is the relative angle in degrees clockwise compared
	 * to the previous pointer_gesture_pinch.begin or
	 * pointer_gesture_pinch.update event.
	 */
	void (*update)(void *data,
		       struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch,
		       uint32_t time,
		       wl_fixed_t dx,
		       wl_fixed_t dy,
		       wl_fixed_t scale,
		       wl_fixed_t rotation);
	/**
	 * end - multi-finger pinch end
	 * @serial: (none)
	 * @time: timestamp with millisecond granularity
	 * @cancelled: 1 if the gesture was cancelled, 0 otherwise
	 *
	 * This event is sent when a multi-finger pinch gesture ceases to
	 * be valid. This may happen when one or more finger is lifted or
	 * the gesture is cancelled.
	 *
	 * When a gesture is cancelled, the client should undo state
	 * changes caused by this gesture. What causes a gesture to be
	 * cancelled is implementation-dependent.
	 */
	void (*end)(void *data,
		    struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch,
		    uint32_t serial,
		    uint32_t time,
		    int32_t cancelled);
};

static inline int
_wl_pointer_gesture_pinch_add_listener(struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch,
				       const struct _wl_pointer_gesture_pinch_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) _wl_pointer_gesture_pinch,
				     (void (**)(void)) listener, data);
}

#define _WL_POINTER_GESTURE_PINCH_DESTROY	0

static inline void
_wl_pointer_gesture_pinch_set_user_data(struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) _wl_pointer_gesture_pinch, user_data);
}

static inline void *
_wl_pointer_gesture_pinch_get_user_data(struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch)
{
	return wl_proxy_get_user_data((struct wl_proxy *) _wl_pointer_gesture_pinch);
}

static inline void
_wl_pointer_gesture_pinch_destroy(struct _wl_pointer_gesture_pinch *_wl_pointer_gesture_pinch)
{
	wl_proxy_marshal((struct wl_proxy *) _wl_pointer_gesture_pinch,
			 _WL_POINTER_GESTURE_PINCH_DESTROY);

	wl_proxy_destroy((struct wl_proxy *) _wl_pointer_gesture_pinch);
}

#ifdef  __cplusplus
}
#endif

#endif
